/* ATTiny2313 MIDI to CV Converter for SX-150             */

#include <avr/io.h>
#include <avr/interrupt.h>

/*---------------------------------------------------------*/

#define CV_PITCH   OCR0A
#define CV_BREATH  OCR1B
#define CV_BEND    OCR1A

#define GATE1		PORTB1	

#define GATE1_ON		(PORTB |= (1 << GATE1))
#define GATE1_OFF		(PORTB &= ~(1 << GATE1))

#define	TRUE			1
#define	FALSE			0
#define ON				1
#define OFF				0

#define	INTRCOUNT		178		// ^C} 250us

#define MIDI_CH 0 /* Receive MIDI Channel */

#ifndef F_CPU
#define F_CPU    20000000	/*  20MHz processor */
#endif

/*---------------------------------------------------------*/
/* set baud rate here */		// for MIDI
#define UART_BAUD_RATE 31250
#define ESC            0x1b
#define UART_BUF_SIZE    16

/* automatically calcuate baud register value */
#define UART_BAUD_SELECT (F_CPU/(UART_BAUD_RATE*16l)-1)

/* MIDI ChannelVoice message */
#define MIDI_NOTE_OFF 0x80
#define MIDI_NOTE_ON  0x90
#define MIDI_P_PLS    0xA0
#define MIDI_C_CHG    0xB0
#define MIDI_P_CHG    0xC0
#define MIDI_C_PLS    0xD0
#define MIDI_PBEND    0xE0

#define CC_BREATH        0x02
#define CC_EXPRESSION    0x0B
#define CC_DAMPER_PEDAL  0x40
#define CC_ALL_SOUND_OFF 0x78
#define CC_ALL_NOTES_OFF 0x7B

#define C0 0x0B

#define true  1
#define false 0
#define ON    1
#define OFF   0

typedef unsigned char  u08;
typedef          char  s08;
typedef unsigned short u16;
typedef          short s16;
typedef u08            bool;

volatile u08 uart_rxd_buf_cnt;
u08 *uart_rxd_in_ptr, *uart_rxd_out_ptr;
u08 uart_rxd_buffer[UART_BUF_SIZE];

volatile uint16_t rxByte;

volatile uint16_t volume;
volatile uint8_t updating;

/* MIDI receiver related */
volatile u08 pc;
volatile u08 sysex;
volatile u08 midi_channel;
volatile u08 mb[2];

/*
 * variables for the note handler
 */
#define MAX_NOTE_COUNT 8

volatile u08  channel_gate;
volatile u08  channel_trig;

volatile uint16_t channel_PitchBend;
volatile u08  channel_Expression;
volatile u08  channel_NotesCount;
volatile u08  channel_Notes[MAX_NOTE_COUNT];
volatile u08  channel_CurrentNote;
volatile u16  channel_CVduty;

/*
 * variables for keyboard reader
 */
volatile uint8_t key_prev;
volatile uint8_t tone_factor;


/**
 * init_uart()
 * Initializes the UART interface
 */
void init_uart(void)
{

  uint16_t baud = UART_BAUD_SELECT;
  UBRRH = (u08)(baud >> 8);
  UBRRL = (u08)baud;
  UCSRB = (1<<RXCIE)|(1<<RXEN);

  uart_rxd_in_ptr  = uart_rxd_out_ptr = uart_rxd_buffer;
  uart_rxd_buf_cnt = 0;
}

void init_timer(void)
{
  // ^C}0ݒ(荞)

    TCCR0A = (1<<COM0A1) /*|(1<<COM0A0)*/|(1<<COM0B1)|(1<<COM0B0)|(1<<WGM01)|(1<<WGM00);
    TCCR0B = (1<<CS00);
    TIMSK = _BV(TOIE0);

    TCCR1A = (1<<COM1A1)|(1<<COM1A0)|(1<<COM1B1)|(1<<COM1B0)| _BV(WGM11) | _BV(WGM10);
    TCCR1B = (1<<WGM12) | _BV(CS10);

    TIMSK |= _BV(TOIE1);

  /**
   * ^C}1ݒ
   * Waveform generation : Fast PWM, 10bit
   * Clock               : No prescaling
   * OC1A initialization : Reset
   * OC1B initialization : Reset
   */
  TCNT1 = 0x0000;
	/*
  TCCR1A = 0xA3;
  TCCR1B = 0x9;
	*/
	// ICR1 = 0xFFF;
    CV_PITCH   = 0xFF;	// pitch CV 
    CV_BREATH  = 0xFF; // volume CV
}

/**
 * ISR()
 * signal handler for receive complete interrupt (RCI)
 */
ISR(USART_RX_vect)
{

  cli();
  *uart_rxd_in_ptr = UDR;	/* read byte from receive register */
  uart_rxd_buf_cnt++;
  if (++uart_rxd_in_ptr >= uart_rxd_buffer + UART_BUF_SIZE) /* Pointer wrapping */
    uart_rxd_in_ptr = uart_rxd_buffer;
  sei();
}

/**
 * uart_getchar()
 * reads UART read buffer and returns a single character.
 */
int16_t uart_getchar(void)
{
  u08 c;

  if (uart_rxd_buf_cnt>0) {
    cli();
    uart_rxd_buf_cnt--;
    c = *uart_rxd_out_ptr;				/* get character from buffer */
    if (++uart_rxd_out_ptr >= uart_rxd_buffer + UART_BUF_SIZE) /* pointer wrapping */
      uart_rxd_out_ptr = uart_rxd_buffer;
    sei();
    return c;
  } else {
    return -1;							/* buffer is empty */
  }
}

void init_MIDI_states()
{
  channel_PitchBend = 8192 >> 4; // 0x20 0x00
//  channel_PitchBend = 10;
  channel_NotesCount = 0;
  channel_CVduty = 0;
  
  channel_gate = 0;
  channel_trig = 0;
}

/**
 * iosetup()
 * Initializes I/O ports.
 */
void iosetup(void)
{
  /* setup output data port */
  DDRB = _BV(PB0) |  // tuning tone
         _BV(PB1) |  // gate out
         _BV(PB2) |  // note CV
         _BV(PB3);   // pitch bend
  //PORTB = 0x00; // turns all output bits on.  "Denji Trigger hardward is designed as logic '0' as switch ON.

    PORTB = 0xf0; // key reader


    PORTD = _BV(PD2) | _BV(PD3) | _BV(PD4) | _BV(PD5) | _BV(PD6);
//    DDRD = (1<<DDD4)|(1<<DDD5)|(1<<DDD6);
}


//-----------------------------------------------
// CVo͓dZbg

void SetCV(void)
{
    uint8_t pwm_value;
    pwm_value = channel_CurrentNote > C0 ? channel_CurrentNote - C0 : 1;
    CV_PITCH = pwm_value * 2;
}

void NoteON(uint8_t note_number);
void NoteOFF(uint8_t note_number);

//-----------------------------------------------
// Set expression CV PWM output

void SetExpression(void)
{
    volume = 1023 - channel_Expression * 8;
}

#define BUZZ_CYCLE (F_CPU / 0xff / 440 / 2)

ISR(TIMER0_OVF_vect)
{
    static uint8_t buzzcount = 0;
    
    if (++buzzcount == BUZZ_CYCLE) {
        buzzcount = 0;
        PORTB ^= tone_factor;
    }
}

ISR(TIMER1_OVF_vect)
{
    static uint16_t button_count = 0;
    uint8_t key;
    uint8_t octave;

    if (CV_BREATH != volume) { 
        if (CV_BREATH > volume) {
            --CV_BREATH;
        }
        else {
    	    ++CV_BREATH;
        }
    }

    if (CV_BEND != channel_PitchBend && ! updating) {
        if (CV_BEND > channel_PitchBend) {
            --CV_BEND;
        }
        else {
            ++CV_BEND;
        }
    }

    // check the keyboard
    //
    octave = ((PIND >> 3) ^ 0x0f) * 12;
    key = ((PINB >> 4) ^ 0x0f);
    if (key) {
        key += octave;
    }
    if (key != key_prev) {
        if (key_prev) {
            NoteOFF(key_prev + C0);
        }
        if (key) {
            NoteON(key + C0);
        }
    }
    key_prev = key;

    if (! (PIND & _BV(PD2))) {
        ++button_count;
    }
    else {
        button_count = 0;
    }
    if (button_count >= 19531) {
        tone_factor ^= _BV(PORTB0);
        button_count = 0;
    }
}

//-----------------------------------------------
// m[gON

void NoteON(uint8_t note_number)
{
  u08 i;
  u08 max = 0;

	// obt@ɓo^ς̃m[g͖
	for(i = 0; i < channel_NotesCount; i++) {
		if(channel_Notes[i] == note_number) {
			return;
		}
	}

	// obt@ςH
	if(channel_NotesCount == MAX_NOTE_COUNT) {
		// ʓ˂
		for(i = 0; i < (MAX_NOTE_COUNT - 1); i++) {
			channel_Notes[i] = channel_Notes[i+1];
		}
		channel_Notes[MAX_NOTE_COUNT - 1] = note_number;
	}
	else {
		channel_Notes[channel_NotesCount] = note_number;
		channel_NotesCount++;
	}

	// ō
	for(i = 0; i < channel_NotesCount; i++) {
		if(max < channel_Notes[i]) {
			max = channel_Notes[i];
		}
	}
	channel_CurrentNote =  max;

	SetCV();
	GATE1_ON;
	channel_gate = 1;
	channel_trig = 1;
}

//-----------------------------------------------
// m[gOFF

void NoteOFF(uint8_t note_number)
{                                                                                          
	u08 i;
	u08 max = 0;
	u08 flg = FALSE;

	// obt@ɓo^Ăꍇ͍폜
	for(i = 0; i < channel_NotesCount; i++) {
		if(flg) {
			channel_Notes[i-1] = channel_Notes[i];
		}
		if(channel_Notes[i] == note_number) {
			flg = TRUE;
		}
	}
	if(flg) channel_NotesCount--;

	if(channel_NotesCount == 0) {
		// obt@J̏ꍇGATEIt
		GATE1_OFF;
		channel_gate = 0;
	}
	else {
		// ō
		for(i = 0; i < channel_NotesCount; i++) {
			if(max < channel_Notes[i]) {
				max = channel_Notes[i];
			}
		}
		channel_CurrentNote = max;
		SetCV();
	}
}

u08 MIDI_SystemMessage(void)
{
  if (sysex) {
    if (rxByte == 0xF7) {
      sysex=0;
    }
  }
  else {
    if (rxByte < 0xF8) {
      if (rxByte > 0x7F) {
        if (rxByte == 0xF0) {
          sysex = 1;
        }
        else {
          midi_channel = rxByte & 0x0F;
        }
        pc=0;
      }
      else {
        mb[pc & 0x01]= rxByte;
      }
      return 1;
    }
  }
  return 0;
}

void MIDI_ChannelMessage(void)
{
    uint16_t temp;

  // Skip process if the channel is not the target.
  if (midi_channel != MIDI_CH) {
    return;
  }

  switch (pc) {
  case 0:
    switch (rxByte & 0xF0) {
    case MIDI_NOTE_OFF : pc=2;  break;
    case MIDI_NOTE_ON  : pc=4;  break;
    case MIDI_P_PLS    : pc=6;  break;
    case MIDI_P_CHG    : pc=8;  break;
    case MIDI_C_CHG    : pc=10; break;
    case MIDI_C_PLS    : pc=12; break;
    case MIDI_PBEND    : pc=14; break;
    }
    break;

  case 2:  /* Note Off */
    pc = 3;
    break;

  case 3:
    pc = 2;
    NoteOFF(mb[0]);
    break;

  case 4: /* Note On */
    pc = 5;
    break;

  case 5:
    pc = 4;
    if (mb[1] == 0) {
      NoteOFF(mb[0]);
    }
    else {
      NoteON(mb[0]);
    }
    break;

  case 6: /* Polyphonic Key Pressure */
    pc = 7;
    break;
  case 7:
    pc = 6;
    break;

  case 8:/* Program change */
    break;

  case 10: /* Control Change */
    pc = 11;
    break;
    
  case 11:
    pc = 10;
    switch(mb[0]) {
		case CC_BREATH:
    case CC_EXPRESSION:
      channel_Expression = mb[1];
      SetExpression();
      break;
    case CC_DAMPER_PEDAL:
    case CC_ALL_SOUND_OFF:
    case CC_ALL_NOTES_OFF:
      pc=0;
      PORTB=0x00;
      break;
    }
    break;

  case 12: /* ChannelPressure */
    break;
    
  case 13:
    break;

  case 14: /* PitchBend */
    pc = 15;
    //channel_PitchBend = mb[0];
    break;
    
  case 15:
    pc = 14;
    temp = 1023 - (mb[0] / 16 + mb[1] * 8);
    updating = 1;
    channel_PitchBend = temp * 5 / 32;
    updating = 0;
    break;

  default:
    break;
  }

}

int main(void)
{
  iosetup();
  init_uart();
  init_timer();
  
  init_MIDI_states();

  pc = 0;
  midi_channel = 0;
  sysex = 0;

  volume = 1023;
  updating = 0;

  key_prev = 0;
  tone_factor = 0;

  sei();

  while(true){
    rxByte = uart_getchar();
    if (rxByte != -1) {
      if(MIDI_SystemMessage()){
        MIDI_ChannelMessage();
      }
    }
  }

}
